import numpy as np
import numexpr as ne
from matplotlib.tri import TriAnalyzer, Triangulation, UniformTriRefiner
from scipy.spatial import Voronoi, ConvexHull

def PDF(pos, nb_samples, rcut, bin_count):
    """
    Pair Distribution Function. Returns normalized histogram of distance between atoms.

    Parameters
    ----------
    pos : np.array
        Array containing atoms position
    nb_samples : int
        Number of atoms from which to generate the histogram
    rcut : number
        Maximum distance to consider
    bin_count : int
        Number of bins of the histogram

    Returns
    -------
    bins, hist : tuple(np.array, np.array)
        `bins` being the distances, `hist` the normalized (regarding radius) histogram

    """
    bins = np.linspace(0, rcut, bin_count)
    samples = np.random.choice(range(len(pos)), nb_samples)
    hist = np.zeros(len(bins)-1)
    for s in samples:
        sample = pos[s,:]
        dists = np.array([a for a in np.sqrt(ne.evaluate("sum((pos-sample)**2,axis=1)")) if a])
        hist += np.histogram(dists, bins=bins, weights=1/dists)[0]
    return bins[:-1], hist/nb_samples


def density(model, refinement=0):
    """
    Create a Voronoi mesh and calculate the local particle density on its vertices.

    The local density is calculated as follows:
    for each vertex, compute the density of each neighbour region as
    one over the area and assign the average of
    the neighbouring density to the vertex.

    Parameters
    ----------
    model : simulation.builder.Model
        the Model object containing
    refinement : int (defaults : 0)
        number of subdivision for refining the mesh (0 == None)
    Returns
    -------
    tri : matplotlib.tri.Triangulation
        the triangulation mesh (refined if set as)
    vert_density : numpy.array
        the array containing the local denstity associated with the tri mesh

    Examples
    -----
    To plot the result using matplotlib use :
    .. code-block:: python
        import matplotlib.pyplot as plt
        tri, density = data_proc.density(model)
        plt.tricontour(tri, density) # to draw contours
        plt.tricontourf(tri, density) # ot draw filled contours
        plt.show()

    Note
    ----
    As of now, the numerical results may not be quantitatively accurate
    but should qualitatively represent the density.
    """
    vor = Voronoi(model.pos)
    vert_density = np.zeros(max(vor.vertices.shape)) # density vector
    reg_num = np.zeros(max(vor.vertices.shape)) # nbr of regions per vertex --> averaging
    for point_index, reg in enumerate(vor.point_region):
        vertices = vor.regions[reg]
        if vertices:
            if -1 not in vertices:
                area = ConvexHull(vor.vertices[vertices]).area # gets the area
                vert_density[vertices] += 1 / area # makes it a density (sort-of)
                reg_num[vertices] += 1
    vert_density /= reg_num # averaging

    # getting rid of really ugly border points
    new_vert, vert_density = (vor.vertices[vor.vertices[:, 0] >= np.min(model.pos[:, 0])],
                              vert_density[vor.vertices[:, 0] >= np.min(model.pos[:, 0])])

    new_vert, vert_density = (new_vert[new_vert[:, 0] <= np.max(model.pos[:, 0])],
                              vert_density[new_vert[:, 0] <= np.max(model.pos[:, 0])])

    new_vert, vert_density = (new_vert[new_vert[:, 1] >= np.min(model.pos[:, 1])],
                              vert_density[new_vert[:, 1] >= np.min(model.pos[:, 1])])

    new_vert, vert_density = (new_vert[new_vert[:, 1] <= np.max(model.pos[:, 1])],
                              vert_density[new_vert[:, 1] <= np.max(model.pos[:, 1])])

    # for triangulation refinement
    tri2 = Triangulation(*new_vert.T)
    if refinement:
        tri2.set_mask(TriAnalyzer(tri2).get_flat_tri_mask(0.1))
        refiner = UniformTriRefiner(tri2)
        print(len(tri2.neighbors), vert_density.shape)
        tri, vert_density = refiner.refine_field(vert_density, subdiv=refinement)
    else:
        tri, vert_density = tri2, vert_density

    return tri, vert_density



"""
Simulator.
Simulates the dynamics of a model (on the CPU or GPU).
"""

import os

from ..utils import gl_util
import moderngl
import numpy as np
import numexpr as ne
import scipy.interpolate as inter
import warnings

from .forces_CPU import ForcesComputeCPU
from .forces_GPU import ForcesComputeGPU

class Simulation:
    """
    Simulator for a model.

    Parameters
    ----------
    model : builder.Model
        Model to simulate.
        The original model object is copied, and thus preserved, which allows it to serve as a reference.

    Attributes
    ----------
    model : builder.Model
        Model that is simulated.

        -------
        Warning
        -------
        Works as usual, but changing its properties will not affect the simulation as intended.
        You should construct another simulation from this model to correctly take in account the changes.
        This is due to the fact that some values (eg. Lennard-Jones parameters) are treated as constants during shader
        initialisation, in order to speed up the calculations.
    current_iter : int
        Number of iterations already computed, since initialisation.
    context : moderngl.Context
        ModernGL context used to build and run compute shader.
    F : np.array
        Last computed forces applied to atoms. Initialized to zeros.

        Warning
        -------
        Changing the values will affect behavior of the model.
    """

    def __init__(self, model = None, simulation = None):

        if simulation:
            model = simulation.model

        self.model = model.copy()

        # Découpage de la liste en segments de taille acceptable par le GPU
        #max_layout_size = gl_util.testMaxSizes()
        max_layout_size = 256 # Probablement optimal (en tout cas d'après essais et guides de bonnes pratiques)
        self.groups_number = int(np.ceil(self.model.npart / max_layout_size))
        self.layout_size = int(np.ceil(self.model.npart / self.groups_number))

        # Chargement et paramétrage du compute shader
        consts = {
            "LAYOUT_SIZE": self.layout_size,
            "X_PERIODIC": 1,
            "Y_PERIODIC": 1,
        }
        for k in model.params:
            consts[k.upper()] = model.params[k]

        try:
            self._compute = ForcesComputeGPU(consts)
        except:
            warnings.warn("GPU not available, falling back on CPU. GPU compute needs OpenGL >=4.3.")
            self._compute = ForcesComputeCPU(consts)

        self.T_f = lambda t:model.T

        if simulation :
            self.current_iter = simulation.current_iter

            self.state_fct = simulation.state_fct

            self.T_cntl = simulation.T_cntl
            if self.T_cntl:
                self.T_f = simulation.T_f

            self.F = simulation.F
        else:
            self.current_iter = 0

            self.state_fct = dict()

            self.state_fct["T"] = []
            self.state_fct["T_ctrl"] = []
            self.state_fct["T_ramps"] = [[],[]]

            self.state_fct["EC"] = []
            self.state_fct["EP"] = []
            self.state_fct["ET"] = []
            self.state_fct["bonds"] = []

            self.state_fct["time"] = []
            self.state_fct["iters"] = []

            self.T_cntl = False

            self.F = np.zeros(self.model.pos.shape) # Doit être initialisé et conservé d'une itération à l'autre

        for s in self.state_fct:
            self.__setattr__(s, self.state_fct[s])

    def iter(self, n=1, callback=None):
        """
        Iterates one or more simulation steps.

        Basically, follows the Position-Verlet method to update positions and speeds, and computes inter-atomic forces
        derived from Lennard-Jones potential.

        Parameters
        ----------
        n: int
            Number of iterations to perform.
        callback : callable
            A callback function that must take the Simulation object as first argument.
            It is called at the end of each iteration.

        Note
        ----
        Setting n is significantly faster than calling :py:meth:`iter` several times.

        Example
        -------
        .. code-block:: python

            model.iter(5)

        Returns
        -------

        """

        for s in self.state_fct:
            self.__setattr__(s, self.state_fct[s])

        betaC = self.T_cntl # Contrôle de la température

        # on crée des alias aux valeurs du modèles pour numexpr
        v = self.model.v
        pos = self.model.pos
        dt = self.model.dt
        dt2 = dt/2.0
        m = self.model.m
        dtm = dt/m
        npart = self.model.npart
        inv2npart = 0.5/npart
        knparts = self.model.kB * npart

        limInf = self.model.lim_inf
        limSup = self.model.lim_sup
        length = self.model.length

        F = self.F
        bondsGL = np.zeros(self.model.npart)

        periodic = self.model.x_periodic or self.model.y_periodic

        if periodic:
            length *= (self.model.x_periodic, self.model.y_periodic)
            # on n'applique les conditions périodiques que selon le(s) axe(s) spécifié(s)

        TVOULUE = 0

        for i in range(n):

            t = self.current_iter * dt # l'heure, qui sert pour le calcul de température

            ne.evaluate("pos + v*dt2", out=pos)  # half drift

            # conditions périodiques de bord
            if periodic:
                ne.evaluate("pos + (pos<limInf)*length - (pos>limSup)*length", out=pos)

            self._compute.set_pos(pos)

            # Énergie cinétique et température
            EC = 0.5 * ne.evaluate("sum(m*v*v)")
            T = EC / knparts
            self.EC.append(EC)
            self.T.append(T)

            F[:] = self._compute.get_F()

            # Énergie potentielle
            EPgl = self._compute.get_PE()
            EP = 0.5 * ne.evaluate("sum(EPgl)")
            self.EP.append(EP)
            self.ET.append(EC + EP)

            # Thermostat
            self.T_ctrl.append(self.T_f(t))
            if betaC:
                beta = np.sqrt(1+self.model.gamma*(self.T_f(t)/T-1))
                ne.evaluate("(v + (F*dtm))*beta", out=v) # kick

            else:
                ne.evaluate("v + (F*dtm)", out=v)  # kick

            ne.evaluate("pos + v*dt2", out=pos)  # half drift

            bondsGL[:] = self._compute.get_COUNT()
            self.bonds.append(inv2npart*ne.evaluate("sum(bondsGL)"))

            self.iters.append(self.current_iter)
            self.time.append(t)

            if callback:
                callback(self)

            self.current_iter += 1

    def set_T_f(self, f):
        """
        Sets function that controls temperature.

        Parameters
        ----------
        f : callable
            Must take time (float) as an argument and return temperature (in K, float).

        Returns
        -------

        """
        self.T_cntl = True
        self.T_f = f
        self.model.T = f(self.current_iter*self.model.dt)

    def set_T_ramps(self, t, T):
        """
        Creates a function based on ramps and uses it for temperature control.
        Values of the function are interpolated between points given in `t` and `T`.
        Temperature is supposed constant before the first point and after the last one.

        Parameters
        ----------
        t : array
            Time.
        T : array
            Associated temperatures.

        Returns
        -------

        """
        f2 = inter.interp1d(t, T)
        self.state_fct["T_ramps"] = [list(t), list(T)]
        self.T_ramps = self.state_fct["T_ramps"]

        def f(x):
            if x<t[0]:
                return T[0]
            elif x>t[-1]:
                return T[-1]
            else:
                return float(f2(x)) # pour la consistance des types
        self.set_T_f(f)


import os, sys
import numpy as np
import json
import datetime
try:
    import fcntl
except ImportError:
    sys.path.append(os.path.dirname(__file__))

from . import datreant as dt
from zipfile import *
from . import appdirs

data_path = appdirs.user_data_dir("open-moldyn")
tmp_path = data_path + "/tmp"
tmp1_path = data_path + "/tmp1"

CATEGORY_LIST = [
    "npart",
    "spc1",
    "spc2",
    "x_a",
    "timestep",
    "border_x_type",
    "border_y_type"
]

class ParamIO(dict):
    """
    An interface to interact with json files as a dictionary.

    Designed to be used with context manager (the with statement)
    and datreant.
    Upon entering the with statement it will attempt to load the json
    file at self.file_name into itself (as a dict).
    When leaving this context, it will store itself into the json file.

    Inherits dict.

    It will try to update the tags and categories of the treant dynState.

    Attributes
    ----------
    dynState : datreant.Treant
        The treant that support the leaf (file) associated with it.
    file_name : datreant.Leaf
        The file name of the json file associated with it.

    Example
    -------
    .. code-block:: python

        t = DynState(dirpath)
        # here t.open returns a ParamIO object as the file is .json
        with t.open("params.json") as IO:
            IO["my_key"] = "my_value"
            random_var = IO[some_key]
        # upon exiting the with block, the content of IO is stored
        # in the .json file
    """

    def __init__(self, dynState : dt.Treant, file: dt.Leaf, **kwargs):
        """

        Parameters
        ----------
        dynState
        file
        kwargs
        """
        super().__init__(**kwargs)
        self.dynState = dynState
        self.file_name = file

    def __enter__(self):
        """
        Try to load the content of the json file into itself

        Try to open the json file from file_name and load the informtions
        it contains into itself.
        Will catch FileNotFoundError and JSONDecodeError

        Returns
        -------
        self : ParamIO
        """
        try:
            with open(self.file_name, mode='r') as file:
                params = json.load(self.file_name)
            for key, value in params.items():
                self[key] = value
        except json.decoder.JSONDecodeError:
            print("File corrupted")
        except FileNotFoundError:
            print("File does not YET exists")

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Open the json file file_name and dump itself into it.

        Open the json file file_name and dump itself into it +
        update the tags and categories of dynState according to
        the CATEGORY_LIST of the module.

        Parameters
        ----------
        exc_type
        exc_val
        exc_tb
        """
        param_exists = False
        try:
            with open(self.file_name, mode='r') as file:
                params = json.load(self.file_name)
                param_exists = True
        except json.decoder.JSONDecodeError:
            print("File corrupted")
        except FileNotFoundError:
            print("File does not YET exists")
        if not param_exists or self != params:
            with open(self.file_name, mode='w') as file:
                json.dump(self, file, ensure_ascii=False, indent=4)
            #self.dynState.categories['last modified'] = datetime.datetime.now().strftime('%d/%m/%Y-%X')
        #self._update_categories()

    def from_dict(self, rdict : dict):
        """
        Copy rdict into itself

        Parameters
        ----------
        rdict : dict
            The remote dictionary from which to copy

        Returns
        -------

        """
        for key, value in rdict.items():
            self[key] = value

    def to_dict(self, rdict : dict):
        """
        Copy itself into the remote dictionary
        Parameters
        ----------
        rdict : dict
            The remot dictionary to which to copy

        Returns
        -------

        """
        for key, value in self.items():
            rdict[key] = value

    def to_attr(self, obj):
        """
        Dump the parameter dictionary in the object (obj) as attributes of said object.

        Warning
        -------
        Will change the value of each attribute of obj that have a name
        corresponding to a key of ParamIO.

        Parameters
        ----------
        obj : an object
            The object to which it will dump its content as attributes

        """
        for key, value in self.items():
            obj.__setattr__(str(key), value)

    def _update_categories(self):
        if self["x_a"] != 0:
            self.dynState.categories["nb_species"] = 2
        for key, value in self.items():
            if key in CATEGORY_LIST:
                self.dynState.categories[key] = value

class NumpyIO:
    """
    An interface to interact with numpy save files with a context manager.

    Attributes
    ----------
    dynState : datreant.Treant
        The treant that support the leaf (file) associated with it.
    mode : str
        The access mode of the file (usually 'r' or 'w' or 'a')
    file_name : datreant.Leaf
        The file name of the .npy file associated with it.
    file : file object
        the file that is opened (contains None until entering a context manager)
    Example
    -------
    .. code-block:: python

        t = DynState(dirpath)
        # here t.open returns a NumpyIO object as the file is .npy
        with t.open("pos.npy", 'r') as IO:
            arr = IO.load() #load an array
        with t.open("pos.npy", 'w') as IO:
            IO.save(arr) #save an array
    """
    def __init__(self, dynState, file : dt.Leaf, mode):
        self.dynState = dynState
        self.mode = mode
        self.file_name = file
        self.file = None

    def __enter__(self):
        self.file = open(self.file_name, mode=self.mode)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        if False and self.file.readable():
            self.dynState.categories['last modified'] = datetime.datetime.now().strftime('%d/%m/%Y-%X')
        """
        self.file.close()

    def save(self, arr):
        """
        Save an array to the opened file.

        Parameters
        ----------
        arr : ndarray (numpy)
            The array to be stored
        Returns
        -------

        """
        np.save(self.file, arr)

    def load(self):
        """
        Load an array stored in the file.
        Returns
        -------
        arr : ndarray
            the array loaded
        """
        return np.load(self.file, allow_pickle=True)


class DynState(dt.Treant):
    """
    A Treant specialized for handling .npy and .json files for moldyn


    """
    POS = "pos.npy" # position of particles
    POS_H = "pos_history.npy" # history of position
    VEL = "velocities.npy" # final velocities
    STATE_FCT = "state_fct.json" # state functions (energy, temperature...)
    PAR = "parameters.json" # parameters of model and simulation

    def __init__(self, treant, *, extraction_path : str = data_path+'/tmp'):
        if isinstance(treant, dt.Treant):
            super().__init__(treant)
        elif isinstance(treant, str) and is_zipfile(treant) :
            try:
                with ZipFile(treant, 'r') as archive:
                    archive.extractall(extraction_path)
                    super().__init__(extraction_path)
            except BadZipFile:
                pass
        else:
            super().__init__(treant)

    def open(self, file, mode='r'):
        if file.endswith(".npy"):
            if not(mode.endswith("+b")):
                mode += "+b"
            return NumpyIO(self, self.leafloc[file], mode)
        elif file.endswith(".json"):
            return ParamIO(self, self.leafloc[file])
        else:
            return open(self.leafloc[file].abspath, mode)

    def add_tag(self,*tags):
        self.tags.add(*tags)

    def to_zip(self, path):
        with ZipFile(path, "w") as archive:
            for leaf in self.leaves():
                if leaf.exists:
                    #print(leaf.relpath)
                    if '/' in leaf.relpath:
                        sep = '/'
                    else:
                        sep = '\\'
                    archive.write(leaf.relpath, arcname=leaf.relpath.split(sep)[-1])

    def save_model(self, model):

        # position of particles
        with self.open(self.POS, 'w') as IO:
            IO.save(model.pos)
        # parameters
        with self.open(self.PAR, 'w') as IO:
            IO.from_dict(model.params)
        # velocity
        with self.open(self.VEL, 'w') as IO:
            IO.save(model.v)


def discover(dirpath=data_path, *args, **kwargs):
    return dt.discover(dirpath=dirpath, *args, **kwargs)



